;; Bootstrap syntax-rules in terms of syntax-case.

(define-syntax syntax-rules
  (lambda (syntax-rules-form)
    (syntax-case syntax-rules-form ()
      ((_ (literal ...) ((keyword . pattern) template) ...)
       #'(lambda (form)
           (syntax-case form (literal ...)
             ((keyword . pattern) #'template) ...))))))

;; Bootstrap define in terms of primitive-define

(define-syntax define
  (syntax-rules ()
    ((_ (name . args) expr expr* ...)
     (primitive-define name (lambda args expr expr* ...)))
    ((_ name expr)
     (primitive-define name expr))))

;; Bootstrap *unspecified*

;; *unspecified* is a constant. It has no lexical syntax, but can be obtained
;; via (if #f #f). It is returned from forms without useful return values.
(define *unspecified* (if #f #f))

;; Bootstrap eqv?

;; Due to the implementation internals, eqv? can be defined as eq?
;; since eq? already compares numeric values.
(define eqv? eq?)

;; Bootstrap list

(define (list . args)
  args)

;; Bootstrap basic macros

(define-syntax begin
  (syntax-rules ()
    ((begin expr ...)
     ;; The *unspecified* serves to allow (begin),
     ;; making it return *unspecified*.
     ((lambda () *unspecified* expr ...)))))

(define-syntax and
  (syntax-rules ()
    ((and) #t)
    ((and x) x) ; to get a tail context
    ((and x y ...)
     (if x (and y ...) #f))))

(define-syntax or
  (syntax-rules ()
    ((or) #f)
    ((or x) x) ; to get a tail context
    ((or x y ...)
     (if x #t (or y ...)))))

(define (not x)
  (if x #f #t))

(define-syntax let
  (syntax-rules ()
    ((let ((var val) ...) let-body let-body* ...)
     ((lambda (var ...) let-body let-body* ...) val ...))
    ((let name ((var val) ...) let-body let-body* ...)
     ((lambda (name)
        (set! name (lambda (var ...) let-body let-body* ...))
        (name val ...))
      *unspecified*))))

(define-syntax cond
  (syntax-rules (=> else)
    ((cond)
     *unspecified*)
    ((cond (test) rest ...)
     (let ((val test))
       (or val (cond rest ...))))
    ((cond (test => fn) rest ...)
     (let ((val test))
       (if val
           (fn val)
           (cond rest ...))))
    ((cond (else expr expr* ...))
     (begin expr expr* ...))
    ((cond (test expr expr* ...) rest ...)
     (if test
         (begin expr expr* ...)
         (cond rest ...)))))

(define-syntax var-case ;; auxiliary macro for `case`
  (syntax-rules (else)
    ((var-case var ((literal ...) expr expr* ...) . rest)
     (if (or (eqv? var 'literal) ...)
         (begin expr expr* ...)
         (var-case var . rest)))
    ((var-case var (else expr expr* ...))
     (begin expr expr* ...))
    ((var-case var)
     *unspecified*)))

(define-syntax case
  (syntax-rules ()
    ((case x . rest)
     (let ((val x))
       (var-case val . rest)))))

(define-syntax when
  (syntax-rules ()
    ((when test when-body when-body* ...)
     (if test (begin when-body when-body* ...)))))

(define-syntax unless
  (syntax-rules ()
    ((unless test unless-body unless-body* ...)
     (if (not test) (begin unless-body unless-body* ...)))))

(define-syntax let*
  (syntax-rules ()
    ((let* () let*-body let*-body* ...)
     (begin let*-body let*-body* ...))
    ((let* ((var val) rest ...) let*-body let*-body* ...)
     (let ((var val)) (let* (rest ...) let*-body let*-body* ...)))))

(define-syntax rec
  (syntax-rules ()
    ((rec name expr)
     (let ((name *unspecified*))
       (set! name expr)
       name))))

(define-syntax letrec
  (syntax-rules ()
    ((letrec ((var val) ...) body body* ...)
     (let ((var *unspecified*) ...)
       (set! var val) ...
       body body* ...))))

;; The only difference between letrec and letrec* is that letrec
;; has an unspecified evaluation order, to allow for fancy optimizations.
;; We don't have any of those, so we can just define letrec* to
;; be the same as letrec.
(define-syntax letrec*
  (syntax-rules ()
    ((letrec* . stuff)
     (letrec . stuff))))

(define-syntax with-syntax
  (syntax-rules ()
    ((with-syntax ((pattern value) ...) body body* ...)
     (syntax-case (list value ...) ()
       ((pattern ...)
        (begin body body* ...))))))

;; TODO: test functions after this point.
;; ======================================

;; Bootstrap basic predicates

(define-syntax define-simple-predicate
  (lambda (sintax)
    (syntax-case sintax ()
      ((define-simple-predicate type)
       (with-syntax ((type? (datum->syntax
                             #'define-simple-predicate
                             (string->symbol
                              (primitive-string-append
                               (symbol->string (syntax->datum #'type))
                               "?")))))
          #'(define (type? x)
              (eq? (value-type x) 'type)))))))

(define-simple-predicate integer)
(define-simple-predicate float)
(define-simple-predicate boolean)
(define-simple-predicate symbol)
(define-simple-predicate null)
(define-simple-predicate unspecified)
(define-simple-predicate procedure)
(define-simple-predicate identifier)
(define-simple-predicate string)
(define-simple-predicate box)
(define-simple-predicate pair)

(define (number? x)
  (or (integer? x) (float? x)))

;; TODO: Use tortoise/hare?
(define (list? x)
  (or (null? x)
      (and (pair? x)
           (list? (cdr x)))))

;; Bootstrap other basic functions

(define (exact>? x y)
  (and (not (exact<? x y))
       (not (eqv? x y))))

(define (exact<=? x y)
  (or (exact<? x y) (eqv? x y)))

(define (exact>=? x y)
  (not (exact<? x y)))

(define (exact- x y)
  (exact+ x (exact-neg y)))

(define (inexact>? x y)
  (and (not (inexact<? x y))
       (not (eqv? x y))))

(define (inexact<=? x y)
  (or (inexact<? x y) (eqv? x y)))

(define (inexact>=? x y)
  (not (inexact<? x y)))

(define (inexact- x y)
  (inexact+ x (inexact-neg y)))

;; Bootstrap list functions

;; TODO: maybe this should detect improper lists.
(define (length lst)
  (if (null? lst)
      0
      (exact+ 1 (length (cdr lst)))))

(define (reverse lst)
  (let loop ((lst lst)
             (acc '()))
    (if (null? lst)
        acc
        (loop (cdr lst)
              (cons (car lst)
                    acc)))))

;; apply is not implemented, so map can only support one list.
(define (map f lst)
  (if (null? lst)
      '()
      (cons (f (car lst))
            (map f (cdr lst)))))

;; same
(define (fold fn init lst)
  (if (null? lst)
      init
      (fold fn (fn (car lst) init) (cdr lst))))

;; Bootstrap string functions

(define (string-append . args)
  (fold primitive-string-append "" (reverse args)))

;; Bootstrap equal?

(define (equal? a b)
  (cond
    ((and (string? a) (string? b))
     (string=? a b))
    ((and (pair? a) (pair? b))
     (and (equal? (car a) (car b))
          (equal? (cdr a) (cdr b))))
    ((and (box? a) (box? b))
     (and (equal? (unbox a)) (equal? (unbox b))))
    (else
     (eqv? a b))))

;; Bootstrap display-to-string and display
(define (display-to-string x)
  (cond
    ((number? x)
     (number->string x))
    ((boolean? x)
     (if x "#t" "#f"))
    ((symbol? x)
     (symbol->string x))
    ((unspecified? x)
     "#<unspecified>")
    ((string? x)
     x)
    ((or (pair? x) (null? x))
     (let loop ((x x)
                (res "(")
                (first #t))
       (cond
         ((null? x)
          (string-append res ")"))
         ((pair? x)
          (loop (cdr x)
                (string-append res
                               (if first "" " ")
                               (display-to-string (car x)))
                 #f))
         (else
          (string-append res " . " (display-to-string x) ")")))))
    ((box? x)
     (string-append "#<box " (display-to-string (unbox x)) ">"))
    ((identifier? x)
     (string-append "#<syntax "
                    (display-to-string (syntax->datum x))
                    ">"))
    ((procedure? x)
     (string-append "#<procedure " (procedure-name x) ">"))))

(define (display x)
  (primitive-display (display-to-string x)))

(define (newline)
  (primitive-display "\n"))

;; Testing helpers

(define-syntax assert
  (syntax-rules ()
    ((assert condition)
     (unless condition
       (display "Assertion failed: ")
       (display 'condition)
       (newline)
       (error "Assertion failure")
       (newline)))))

(define-syntax test-group
  (syntax-rules ()
    ((test-group condition condition* ...)
     (begin
      (assert condition)
      (assert condition*) ...
      (display "Test execution terminated successfully.\n")))))
