;;; This is an implementation of the Unlambda programming language,
;;; ported to kscheme from the original.

;;; Copyright (C) 1999 by David A. Madore <david.madore@ens.fr>
;;; Copyright (C) 2023 by Jean Abou Samra <jean@abou-samra.fr>

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version
;;; 2 of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty
;;; of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
;;; the GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

;; Data representation: f applied to g is represented as (f . g) (we
;; use a pair rather than a list because application *always* takes
;; *exactly* one argument).  Primitive procedures k, s, and so on are
;; represented as (k), (s) and so on.  Derived procedures (k1, s1, s2
;; and the like) are represented by lists whose first element is k1,
;; s1... and whose subsequent elements give internal data to the
;; derived procedure (data that was set there by whatever created the
;; procedure, to be remember when it is called).

;; I would be very grateful if somebody could explain to me how all
;; this works.  -- The author


;; A port is represented as a boxed list of single-character strings
;; (this implementation doesn't have a character type, so we use
;; single-character strings instead).

(define (string->list str)
  (let loop ((n 0) (acc '()))
    (if (eqv? n (string-length str))
        (reverse acc)
        (loop (exact+ n 1)
              (cons (substring str n (exact+ n 1)) acc)))))

(define (make-port-from-string str)
  (box (string->list str)))

;; We use *unspecified* for the EOF object. (Standard Scheme
;; has a dedicated object, but it is not implemented here.)
(define (read-char port)
  (let ((char-list (unbox port)))
    (if (pair? char-list)
        (let ((ch (car char-list)))
          (set-box! port (cdr char-list))
          ch)
        *unspecified*)))

;; Parse the input file and return the representation of it.
(define (read-char-or-error input-port)
  (let ((ch (read-char input-port)))
    (if (unspecified? ch)
        (error "Unexpected end of file")
        ch)))

(define (gobble-comment input-port)
  (let ((ch (read-char input-port)))
    (unless (or (unspecified? ch)
            	(equal? ch "\n"))
      (gobble-comment input-port))))

(define (parse input-port)
  (let ((ch (read-char-or-error input-port)))
    (cond
     ((equal? ch "`")
      (let* ((op (parse input-port))
	           (arg (parse input-port)))
        (cons op arg)))
     ((or (equal? ch " ")
          (equal? ch "\n"))
      (parse input-port))
     ((equal? ch "#")
      (gobble-comment input-port)
      (parse input-port))
     ((or (equal? ch "k") (equal? ch "K")) ; (lambda (x) (lambda (y) x))
      '(k))
     ((or (equal? ch "s") (equal? ch "S")) ; (lambda (x) (lambda (y) (lambda (z) ((x z) (y z)))))
      '(s))
     ((or (equal? ch "i") (equal? ch "I")) ; (lambda (x) x)
      '(i))
     ((or (equal? ch "d") (equal? ch "D")) ; delay
      '(d))
     ((equal? ch "r")
      '(pr "\n"))
     ((equal? ch ".")
      (list 'pr (read-char-or-error input-port)))
     (else
      (error (string-append "Character not understood: " ch))))))

;; kscheme does not support mutually recursive definitions
;; at the toplevel. Work around using set!.
(define ap #f)

;; The eval function
(define (ev exp)
  (if (pair? (car exp))
      (let ((op (ev (car exp))))
        (if (eq? (car op) 'd)
	    (list 'd1 (cdr exp))
	    (ap op (ev (cdr exp)))))
      exp))

(define (first lst) (car lst))
(define (second lst) (car (cdr lst)))
(define (third lst) (car (cdr (cdr lst))))

;; The apply function
(set! ap
      (lambda (exp arg)
        (case (first exp)
          ((k)
           (list 'k1 arg))
          ((k1)
           (second exp))
          ((s)
           (list 's1 arg))
          ((s1)
           (list 's2 (second exp) arg))
          ((s2)
           (ev (cons (cons (second exp) arg)
                     (cons (third exp) arg))))
          ((i)
           arg)
          ((pr)
           (display (second exp))
           arg)
          ((d1)
           (ev (cons (second exp) arg)))
          (else
           (error "Internal error: unexpected type to apply!")))))

(ev (parse (make-port-from-string (read-stdin))))
